require 'sinatra/base'
require 'uu/os/qos/qos_handler'
require 'uu/os/rest/remote_client'
require 'uu/os/security/session'
require 'uu/os/vuc/http_status'
require 'uu/os/vuc/view'
require 'uu_adk_client/version'
require 'uu_os_gvc/version'
require 'uu_os_client/version'
require 'uu_os_framework/version'
require 'uu/os/logger'
require 'uu/os/vuc/logger_context'
require 'uu/os/vuc/use_case_log_writer'

module UU
  module OS
    module VUC

      # Ancestor for development of custom form controllers.
      # All controllers should define only event handling methods.
      #
      # @example Event method definition:
      #           def on_<%event_code%>(view)
      #             # 1. process and modify view
      #             # 2. return status from {UU::ADK::HTTPStatus}
      #             #    (optional, 200/OK is returned if not defined)
      #           end
      class FormController < Sinatra::Base

        post '/*' do
          do_post(request.base_url, env['HTTP_AUTHORIZATION'], request.body.read)
        end

        get '/version' do
          erb :version, :locals => {
            :title => self.class,
            :uu_adk_client__version => UUADKClient::VERSION,
            :uu_os_gvc__version => UUOSGVC::VERSION,
            :uu_os_client__version => UUOSClient::VERSION,
            :uu_os_framework__version => UUOSFramework::VERSION
          }
        end

        template :version do
          template = <<-EOS
          <html>
            <head>
              <title><%=title%></title>
            </head>
            <body>
              <b><code>uu_adk_client.gem</code> version:</b> <%= uu_adk_client__version %><br>
              <b><code>uu_os_gvc.gem</code> version:</b> <%= uu_os_gvc__version %><br>
              <b><code>uu_os_client.gem</code> version:</b> <%= uu_os_client__version %><br>
              <b><code>uu_os_framework.gem</code> version:</b> <%= uu_os_framework__version %><br>
            </body>
          </html>
          EOS
        end

        # Creates new instance of FormController.
        def initialize
          super
          @log = UU::OS::Logger.new(self.class)
        end

        # Controller need not to implement all possible event methods. This method
        # provides default behavior of all unimplemented events.
        #
        # @private
        def method_missing(method, *args, &block)
          view = args[0]
          # Do not touch view in case of unimplemented event with exception
          # of submit event which must be always implemented.
          if (method == :on_submit)
            raise "Controller #{self.class} does not define on_submit method."
          elsif (method.to_s[/^on_/] == nil) || (args.size != 1)
            # Invoked method was not "event" method with view parameter - throw regular method missing exception.
            raise NameError.new("undefined local variable or method '#{method.to_s}' for #{self.class}")
          end
          return HTTPStatus::OK
        end

        private

        # Dispatches functionality of controller.
        #
        # @param [String] base_url Base URL on which was controller called
        # @param [String] auth_token Authentication token to be used by default
        # @param [String] body Request body containing view data
        #
        # @return [Array<Numeric, String>] Array containing HTTP status code (should be one of {UU::OS::VUC::HTTPStatus}) as first element and response body as second element
        #
        # @private
        def do_post(base_url, auth_token, body)
          begin
            # Controllers should respond as fast as possible, they should not wait on QoS
            UU::OS::QoS::QoSHandler.default_max_client_timeout = 0
            # If not configured, call all services on same base URL as this controller
            UU::OS::REST::RemoteClient.default_base_url = base_url
            # Parse incoming request and create view model
            view = UU::OS::VUC::View.new(body)
            # Set diagnostic information to logger context
            UU::OS::VUC::LoggerContext.set(view.context)
            # If authentication data are provided, authenticate user
            # (Authentication is called before every other operation to be sure no
            # processing is done without valid authentication data)
            UU::OS::Security::Session.login(auth_token) if auth_token
            # Invoke controller method according to incomming event
            ctrl_method = get_event_method(view.event.code)
            do_log = (ctrl_method =~ /^on_init$/)
            if do_log
              log_params = {}
              log_params[:start] = Time.now
              log_params[:level] = UU::OS::UseCaseLog::UseCaseLogLevel::ERROR
              log_params[:type] = UU::OS::UseCaseLog::UseCaseLogType::VISUAL_USE_CASE
              log_params[:session] = UU::OS::Security::Session.current_session
              log_params[:view] = view
            end
            self.send(ctrl_method, view)
            status = view.status
            if (status.nil?) || (!status.kind_of?Numeric)
              # Invoke method did not provide status code or
              # returned some other object than status number
              status = HTTPStatus::OK
            end
            if do_log
              log_params[:level] = UU::OS::UseCaseLog::UseCaseLogLevel::INFO if status.to_i < 400
            end
            resp_body = view.to_response
            log_request(ctrl_method, body, resp_body)
            [status, resp_body]
          rescue Exception => e
            log_params[:message] = e.to_s if do_log
            @log.error('Execution of controller failed.', e)
            [HTTPStatus::INTERNAL_SERVER_ERROR, get_error_response(e)]
          ensure
            if do_log
              log_params[:finish] = Time.now
              UU::OS::VUC::UseCaseLogWriter.add_record(log_params)
            end
            # Always do complete deletion of all tokens
            UU::OS::Security::Session.logout(true)
            # Always clean up all diagnostic information from logger context
            UU::OS::VUC::LoggerContext.clean
          end
        end

        # Returns name of method for handling of event
        # (transforms camel case to snake case).
        #
        # @param [String] event Event to be handled
        #
        # @return [Symbol] Name of method for handling event
        #
        # @private
        def get_event_method(event)
          # Based on ActiveSupport::Inflector.underscore
          # TODO This method is also on AttachmentDispatcher. We need some "util" class/module to share this code
          result = event.to_s.dup
          result.gsub!(/([A-Z]+)([A-Z][a-z])/,'\1_\2')
          result.gsub!(/([a-z\d])([A-Z])/,'\1_\2')
          result.tr!("-", "_")
          ('on_' << result.downcase).to_sym
        end

        # Returns form error response.
        #
        # @param [Exception] exception Instance of exception
        #
        # @return [String] Error response
        #
        # @private
        def get_error_response(exception)
          # TODO This method is also on AttachmentDispatcher. We need some "util" class/module to share this code
          id = (exception.respond_to?:id) ? exception.id : ''
          code = (exception.respond_to?:code) ? exception.code : ''
          message = (exception.respond_to?:message) ? exception.message : ''
          "{\"id\":\"#{id}\",\"code\":\"#{code}\",\"message\":\"#{message}\",\"time\":\"#{Time.now}\"}"
        end

        # Logs form request
        #
        # @param [String] method Invoked method
        # @param [String] request Incoming request
        # @param [String] response Outgoing response
        #
        # @private
        def log_request(method, request, response)
          if @log.debug_loggable?
            req = (request.nil?) ? request : request.dup.force_encoding('UTF-8')
            resp = (response.nil?) ? response : response.dup.force_encoding('UTF-8')
            @log.debug("Request processed.\n\tController: #{self.class}\n\tMethod: #{method}\n\tRequest: #{req}\n\tResponse:#{resp}")
          end
        end

      end

    end
  end
end
